import asyncio
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Optional

from ..pool import ConnectionPool, ClosedPool, EmptyPool

from .aioconnection import AIOLDAPConnection

MYPY = False

if MYPY:
    from ..ldapclient import LDAPClient


class AIOConnectionPool(ConnectionPool[AIOLDAPConnection]):
    """
    A connection pool that can be used with asnycio tasks. It's inherited from
    :class:`bonsai.pool.ConnectionPool`.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param \\*\\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """

    def __init__(
        self,
        client: "LDAPClient",
        minconn: int = 1,
        maxconn: int = 10,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        **kwargs: Any
    ):
        super().__init__(client, minconn, maxconn, **kwargs)
        self._loop = loop
        try:
            # The loop parameter is deprecated since 3.8, removed in 3.10
            # and it raises TypeError.
            self._lock = asyncio.Condition(loop=self._loop)
        except TypeError:
            self._lock = asyncio.Condition()

    async def open(self) -> None:
        async with self._lock:
            for _ in range(
                self._minconn - self.idle_connection - self.shared_connection
            ):
                conn = await self._client.connect(
                    is_async=True, loop=self._loop, **self._kwargs
                )
                self._idles.add(conn)
            self._closed = False

    async def get(self) -> AIOLDAPConnection:
        async with self._lock:
            if self._closed:
                raise ClosedPool("The pool is closed.")
            await self._lock.wait_for(lambda: not self.empty or self._closed)
            try:
                conn = self._idles.pop()
            except KeyError:
                if len(self._used) < self._maxconn:
                    conn = await self._client.connect(
                        is_async=True, loop=self._loop, **self._kwargs
                    )
                else:
                    raise EmptyPool("Pool is empty.") from None
            self._used.add(conn)
            self._lock.notify()
            return conn

    async def put(self, conn: AIOLDAPConnection) -> None:
        async with self._lock:
            super().put(conn)
            self._lock.notify()

    async def close(self) -> None:
        async with self._lock:
            super().close()
            self._lock.notify_all()

    @asynccontextmanager
    async def spawn(
        self, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[AIOLDAPConnection, None]:
        conn = None
        try:
            if self._closed:
                await self.open()
            conn = await self.get(*args, **kwargs)
            yield conn
        finally:
            if conn:
                await self.put(conn)



import pytest

from bonsai.active_directory import SID


def test_create_from_bytes_le():
    """Test create SID from binary."""
    with pytest.raises(ValueError):
        _ = SID(bytes_le=b"NOT-A-SID")
    with pytest.raises(ValueError):
        _ = SID(
            bytes_le=b"\x01\x05\x00\x00\x00\x05\x15\x00\x00\x00\xdc\xf4\xdc;\x83=+F\x82\xa6(\x00\x02\x00\x00"
        )
    with pytest.raises(TypeError):
        _ = SID(bytes_le=True)
    with pytest.raises(TypeError):
        _ = SID(bytes_le="S-1-1-1")
    input_data = b"\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xdc\xf4\xdc;\x83=+F\x82\x8b\xa6(\x00\x02\x00\x00"
    expected_sid = "S-1-5-21-1004336348-1177238915-682003330-512"
    sid = SID(bytes_le=input_data)
    assert sid.revision == 1
    assert sid.identifier_authority == 5
    assert sid.subauthorities == (21, 1004336348, 1177238915, 682003330, 512)
    assert sid == expected_sid


def test_create_from_str_rep():
    """Test create SID from string."""
    with pytest.raises(ValueError):
        _ = SID("NOT-A-SID")
    with pytest.raises(ValueError):
        _ = SID("S--")
    with pytest.raises(ValueError):
        _ = SID(str_rep="S-1")
    with pytest.raises(ValueError):
        _ = SID("S-1-NOT-SID")
    with pytest.raises(ValueError):
        _ = SID("NOT")
    with pytest.raises(TypeError):
        _ = SID(str_rep=0)
    with pytest.raises(TypeError):
        _ = SID(b"S-1-1-1")
    sid = SID("S-1-5-21-3623811015-3361044348-30300820-1013")
    assert sid.revision == 1
    assert sid.identifier_authority == 5
    assert sid.subauthorities == (21, 3623811015, 3361044348, 30300820, 1013)
    everyone_sid = SID("S-1-1-0")
    assert everyone_sid.identifier_authority == 1
    assert everyone_sid.subauthorities == (0,)


def test_bytes_le():
    """Test bytes_le property."""
    expected_bytes = b"\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xdc\xf4\xdc;\x83=+F\x82\x8b\xa6(\x00\x02\x00\x00"
    test_sid = SID("S-1-5-21-1004336348-1177238915-682003330-512")
    assert test_sid.bytes_le == expected_bytes
    assert (
        SID(bytes_le=test_sid.bytes_le).identifier_authority
        == test_sid.identifier_authority
    )
    assert SID(bytes_le=test_sid.bytes_le).subauthorities == test_sid.subauthorities


def test_str():
    """Test __str__ method."""
    test_sid = "S-1-5-21-3623811015-3361044348-30300820-1013"
    assert str(SID(test_sid)) == test_sid
    test_sid = "S-1-1-0"
    assert str(SID(test_sid)) == test_sid


def test_repr():
    """Test __repr__ method."""
    sid = SID("S-1-1-0")
    assert SID.__name__ in repr(sid)
    assert "S-1-1-0" in repr(sid)


def test_eq():
    """Test __eq__ method."""
    sid0 = SID("S-1-1-0")
    sid1 = SID("S-1-1-0")
    assert sid0 == sid1
    assert sid0 != True
    assert sid0 == "S-1-1-0"


def test_hash():
    """Test __hash__ method."""
    test_sid = "S-1-5-21-3623811015-3361044348-30300820-1013"
    sid0 = SID(test_sid)
    assert hash(sid0) == hash(test_sid)
    assert set((sid0, SID(test_sid))) == {sid0}


@pytest.mark.parametrize(
    "sid, alias",
    [
        (SID("S-1-1-0"), "WD"),
        (SID("S-1-5-32-559"), "LU"),
        (SID("S-1-5-21-3623811015-3361044348-30300820-500"), "LA"),
        (SID("S-1-5-21-3623811015-3361044348-30300820-520"), "PA"),
        (SID("S-1-5-21-3623811015-3361044348-30300820-1013"), None),
        (SID("S-1-5-34-3623811015-3361044348-30300820-500"), None),
    ],
)
def test_sddl_alias(sid, alias):
    """Test sddl_alias property."""
    assert sid.sddl_alias == alias


@pytest.mark.parametrize(
    "sid",
    [
        SID("S-1-1-0"),
        SID("S-1-5-32-559"),
        SID("S-1-5-34-3623811015-3361044348-30300820-500"),
    ],
)
def test_size(sid):
    """Test size property."""
    assert sid.size == len(sid.bytes_le)



import os
import subprocess
import xmlrpc.server as rpc
import time
import sys
import multiprocessing as mp

try:
    import pydivert
except ImportError:
    pass


class LinuxDelayHandler:
    @staticmethod
    def get_interface_name():
        """ Get the first interface name that is not the localhost. """
        net = os.listdir("/sys/class/net")
        net.remove("lo")
        if "eth0" in net:
            return "eth0"
        return net[0]

    def set_delay(self, sec, duration=10.0):
        """ Set network delay, return with the call's result. """
        try:
            subprocess.check_call(
                [
                    "tc",
                    "qdisc",
                    "add",
                    "dev",
                    self.get_interface_name(),
                    "root",
                    "handle",
                    "1:",
                    "prio",
                ]
            )
            subprocess.check_call(
                [
                    "tc",
                    "qdisc",
                    "add",
                    "dev",
                    self.get_interface_name(),
                    "parent",
                    "1:3",
                    "handle",
                    "30:",
                    "netem",
                    "delay",
                    ("%dmsec" % (sec * 1000)),
                ]
            )
            for port in ("389", "636"):
                subprocess.check_call(
                    [
                        "tc",
                        "filter",
                        "add",
                        "dev",
                        self.get_interface_name(),
                        "protocol",
                        "ip",
                        "parent",
                        "1:0",
                        "u32",
                        "match",
                        "ip",
                        "sport",
                        port,
                        "0xffff",
                        "flowid",
                        "1:3",
                    ]
                )

            return True
        except subprocess.CalledProcessError:
            return False

    def remove_delay(self):
        """ Remove network delay. """
        try:
            subprocess.check_call(
                ["tc", "qdisc", "del", "dev", self.get_interface_name(), "root"]
            )
            return True
        except subprocess.CalledProcessError:
            return False


class MacDelayHandler:
    def set_delay(self, sec, duration=10.0):
        with open("/etc/pf.conf") as fp:
            conf = fp.read()
            conf += '\ndummynet-anchor "mop"\nanchor "mop"\n'
            rule = (
                "dummynet in quick proto tcp from any to any port {389, 636} pipe 1\n"
            )

            try:
                subprocess.run(
                    ["pfctl", "-f", "-"], input=conf, encoding="utf-8", check=True
                )
                subprocess.run(
                    ["pfctl", "-a", "mop", "-f", "-"],
                    input=rule,
                    encoding="utf-8",
                    check=True,
                )
                subprocess.check_call(
                    ["dnctl", "pipe", "1", "config", "delay", "%d" % int(sec * 1000)]
                )

                return True
            except subprocess.CalledProcessError:
                return False

    def remove_delay(self):
        try:
            subprocess.check_call(["dnctl", "-q", "flush"])
            subprocess.check_call(["pfctl", "-f", "/etc/pf.conf"])
            return True
        except subprocess.CalledProcessError:
            return False


class WinDelayHandler:
    proc = None

    def delay(self, sec, duration=10.0):
        netfil = "tcp.DstPort == 389 or tcp.SrcPort == 389"
        start = time.time()
        with pydivert.WinDivert(netfil) as divert:
            for packet in divert:
                time.sleep(sec)
                divert.send(packet)
                if time.time() - start >= duration:
                    break

    def set_delay(self, sec, duration=10.0):
        """ Set network delay, return with the call's result. """
        self.proc = mp.Process(target=self.delay, args=(sec, duration))
        self.proc.start()
        return True

    def remove_delay(self):
        """ Remove network delay, return with the call's result. """
        if self.proc is not None and self.proc.is_alive():
            self.proc.terminate()
        return True


if __name__ == "__main__":
    if sys.platform == "win32":
        handler = WinDelayHandler()
    elif sys.platform == "darwin":
        handler = MacDelayHandler()
    else:
        handler = LinuxDelayHandler()
        # Fix network collapse on certain Linux distros.
        subprocess.call(
            ["ip", "link", "set", handler.get_interface_name(), "qlen", "1000"]
        )

    server = rpc.SimpleXMLRPCServer(("0.0.0.0", 8000))
    server.register_function(handler.set_delay, "set_delay")
    server.register_function(handler.remove_delay, "remove_delay")
    server.serve_forever()



import struct

from typing import Dict, Optional, Union

from .sid import SID
from .acl import ACL, ACLRevision, ACE, ACEFlag, ACERight, ACEType


class SecurityDescriptor:
    """
    A class that defines the security attributes of an object. These
    attributes specify who owns the object; who can access the object and
    what they can do with it; what level of audit logging can be applied to
    the object; and what kind of restrictions apply to the use of the
    security descriptor. It can contain two types of :class:`ACL`:

    * A discretionary access control list (DACL) is controlled by the
      owner of an object or anyone granted WRITE_DAC access to the object.
      It specifies the access particular users and groups can have to an
      object. For example, the owner of a file can use a DACL to control
      which users and groups can and cannot have access to the file.
    * A system access control list (SACL) is similar to the DACL, except
      that the SACL is used to audit rather than control access to an
      object. When an audited action occurs, the operating system records
      the event in the security log. Each ACE in a SACL has a header that
      indicates whether auditing is triggered by success, failure, or
      both; a SID that specifies a particular user or security group to
      monitor; and an access mask that lists the operations to audit.

    :param dict control: a dict that specifies control access bit flags.
    :param SID|None owner_sid: the SID of the owner of the object.
    :param SID|None group_sid: the SID of the group of the object.
    :param ACL|None sacl: the system access control list.
    :param ACL|None dacl: the discretionary access control list.
    :param int revision: the revision of the security descriptor.
    :param int sbz1: reserved value.
    """

    def __init__(
        self,
        control: Dict[str, bool],
        owner_sid: Optional[SID],
        group_sid: Optional[SID],
        sacl: Optional[ACL],
        dacl: Optional[ACL],
        revision: int = 1,
        sbz1: int = 0,
    ) -> None:
        self.__revision = revision
        self.__sbz1 = sbz1
        self.__control = control
        self.__owner_sid = owner_sid
        self.__group_sid = group_sid
        self.__sacl = sacl
        self.__dacl = dacl

    @staticmethod
    def __convert_ctrl(ctrl, to_int=False):
        values = {
            "self_relative": 0x8000,
            "rm_control_valid": 0x4000,
            "sacl_protected": 0x2000,
            "dacl_protected": 0x1000,
            "sacl_auto_inherited": 0x800,
            "dacl_auto_inherited": 0x400,
            "sacl_computed_inheritance_required": 0x200,
            "dacl_computed_inheritance_required": 0x100,
            "server_security": 0x80,
            "dacl_trusted": 0x40,
            "sacl_defaulted": 0x20,
            "sacl_present": 0x10,
            "dacl_defaulted": 0x8,
            "dacl_present": 0x4,
            "group_defaulted": 0x2,
            "owner_defaulted": 0x1,
        }
        if to_int:
            return sum(values[key] for key, val in ctrl.items() if val)
        else:
            return {key: bool(ctrl & val) for key, val in values.items()}

    @classmethod
    def from_binary(cls, data: bytes) -> "SecurityDescriptor":
        """
        Create a SecurityDescriptor object from a binary blob.

        :param bytes data: a little-endian byte ordered byte input.
        :returns: A new SecurityDescriptor instance.
        :rtype: SecurityDescriptor
        :raises TypeError: when the parameter is not bytes.
        :raises ValueError: when the input cannot be parsed as a
            SecurityDescriptor object.
        """
        try:
            if not isinstance(data, bytes):
                raise TypeError("The `data` parameter must be bytes")
            sacl = None
            dacl = None
            (
                rev,
                sbz1,
                ctrl,
                offset_owner,
                offset_group,
                offset_sacl,
                offset_dacl,
            ) = struct.unpack("<BBHIIII", data[:20])
            ctrl = cls.__convert_ctrl(ctrl)
            owner_sid = SID(bytes_le=data[offset_owner:]) if offset_owner else None
            group_sid = SID(bytes_le=data[offset_group:]) if offset_group else None
            if ctrl["sacl_present"] and offset_sacl != 0:
                sacl = ACL.from_binary(data[offset_sacl:])
            if ctrl["dacl_present"] and offset_dacl != 0:
                dacl = ACL.from_binary(data[offset_dacl:])
            return cls(ctrl, owner_sid, group_sid, sacl, dacl, rev, sbz1)
        except struct.error as err:
            raise ValueError(f"Not a valid binary SecurityDescriptor, {err}")

    def to_binary(self) -> bytes:
        """
        Convert SecurityDescriptor object to binary form with little-endian byte order.

        :returns: Bytes of the binary SecurityDescriptor instance
        :rtype: bytes
        """
        owner = b""
        group = b""
        sacl = b""
        dacl = b""
        offset_owner = 0
        offset_group = 0
        offset_sacl = 0
        offset_dacl = 0
        ctrl = self.__convert_ctrl(self.control, True)
        data = bytearray(20)
        if self.owner_sid:
            owner = self.owner_sid.bytes_le
            offset_owner = 20
        if self.group_sid:
            group = self.group_sid.bytes_le
            offset_group = 20 + len(owner)
        if self.sacl:
            sacl = self.sacl.to_binary()
            offset_sacl = 20 + len(owner) + len(group)
        if self.dacl:
            dacl = self.dacl.to_binary()
            offset_dacl = 20 + len(owner) + len(group) + len(sacl)
        struct.pack_into(
            "<BBHIIII",
            data,
            0,
            self.revision,
            self.sbz1,
            ctrl,
            offset_owner,
            offset_group,
            offset_sacl,
            offset_dacl,
        )
        data.extend(owner)
        data.extend(group)
        data.extend(sacl)
        data.extend(dacl)
        return bytes(data)

    def set_control(self, control: Union[Dict[str, bool], int]) -> None:
        """
        Set the control access bit flags for the SecurityDescriptor object.

        :param dict|int control: the control flag either as an int or a dict.
        :raises ValueError: if the provided number is not valid.
        :raises TypeError: if the parameter neither an int nor a bool value \
        only dict.
        """
        if isinstance(control, int):
            ctrl = self.__convert_ctrl(control)
            if control != self.__convert_ctrl(ctrl, to_int=True):
                raise ValueError(f"{control} is not a valid controll flag.")
            self.__control = ctrl
        elif isinstance(control, dict) and all(
            isinstance(val, bool) for val in control.values()
        ):
            self.__control = control
        else:
            raise TypeError(
                "The control parameter must be an int or bool value only dictionary."
            )

    def set_owner_sid(self, owner_sid: SID) -> None:
        """
        Set the owner SID for the the SecurityDescriptor object.

        :param SID owner_sid: the new owner SID.
        :raises TypeError: if the parameter is not a SID object.
        """
        if not isinstance(owner_sid, SID):
            raise TypeError("The owner_sid must be a SID object.")
        self.__owner_sid = owner_sid

    def set_group_sid(self, group_sid: SID) -> None:
        """
        Set the group SID for the the SecurityDescriptor object.

        :param SID group_sid: the new group SID.
        :raises TypeError: if the parameter is not a SID object.
        """
        if not isinstance(group_sid, SID):
            raise TypeError("The group_sid must be a SID object.")
        self.__group_sid = group_sid

    def set_dacl(self, dacl: Optional[ACL]) -> None:
        """
        Set the discretionary ACL for the the SecurityDescriptor object.
        It also sets the `dacl_present` in the control access bit flags
        accordingly.

        :param ACL|None dacl: the ACL object or None.
        :raises TypeError: if the parameter is not a ACL object or None.
        """
        if dacl is not None and not isinstance(dacl, ACL):
            raise TypeError("The dacl must be an ACL object or None.")
        self.__dacl = dacl
        if dacl is not None:
            self.__control["dacl_present"] = True
        else:
            self.__control["dacl_present"] = False

    def set_sacl(self, sacl: Optional[ACL]) -> None:
        """
        Set the system ACL for the the SecurityDescriptor object.
        It also sets the `sacl_present` in the control access bit flags
        accordingly.

        :param ACL|None dacl: the ACL object or None.
        :raises TypeError: if the parameter is not a ACL object or None.
        """
        if sacl is not None and not isinstance(sacl, ACL):
            raise TypeError("The sacl must be an ACL object or None.")
        self.__sacl = sacl
        if sacl is not None:
            self.__control["sacl_present"] = True
        else:
            self.__control["sacl_present"] = False

    @property
    def sbz1(self) -> int:
        """Reserved field in the security descriptor structure."""
        return self.__sbz1

    @property
    def revision(self) -> int:
        """The revision of the security descriptor."""
        return self.__revision

    @property
    def control(self) -> Dict[str, bool]:
        """The dict of the control access bit flags."""
        return self.__control

    @control.setter
    def control(self, value: Union[Dict[str, bool], int]) -> None:
        self.set_control(value)

    @property
    def owner_sid(self) -> Optional[SID]:
        """The :class:`SID` of the owner."""
        return self.__owner_sid

    @owner_sid.setter
    def owner_sid(self, value: SID) -> None:
        self.set_owner_sid(value)

    @property
    def group_sid(self) -> Optional[SID]:
        """The :class:`SID` of the group."""
        return self.__group_sid

    @group_sid.setter
    def group_sid(self, value: SID) -> None:
        self.set_group_sid(value)

    @property
    def sacl(self) -> Optional[ACL]:
        """The system :class:`ACL`."""
        return self.__sacl

    @sacl.setter
    def sacl(self, value: Optional[ACL]) -> None:
        self.set_sacl(value)

    @property
    def dacl(self) -> Optional[ACL]:
        """The discretionary :class:`ACL`."""
        return self.__dacl

    @dacl.setter
    def dacl(self, value: Optional[ACL]) -> None:
        self.set_dacl(value)


class UserAccountControl:
    """
    A class for parsing UserAccountControl field.

    :param int flags: integer representing the property flags.
    :raises TypeError: if flags parameter is not an int.
    """

    def __init__(self, flags: int) -> None:
        if not isinstance(flags, int):
            raise TypeError("The `flags` parameter must be an integer")
        self.__flag_values = {
            "script": 0x1,
            "accountdisable": 0x2,
            "homedir_required": 0x8,
            "lockout": 0x10,
            "passwd_notreqd": 0x20,
            "passwd_cant_change": 0x40,
            "encrypted_text_pwd_allowed": 0x80,
            "temp_duplicate_account": 0x100,
            "normal_account": 0x200,
            "interdomain_trust_account": 0x800,
            "workstation_trust_account": 0x1000,
            "server_trust_account": 0x2000,
            "dont_expire_password": 0x10000,
            "mns_logon_account": 0x20000,
            "smartcard_required": 0x40000,
            "trusted_for_delegation": 0x80000,
            "not_delegated": 0x100000,
            "use_des_key_only": 0x200000,
            "dont_req_preauth": 0x400000,
            "password_expired": 0x800000,
            "trusted_to_auth_for_delegation": 0x1000000,
            "partial_secrets_account": 0x4000000,
        }
        self.__properties = {
            key: bool(flags & val) for key, val in self.__flag_values.items()
        }

    @property
    def properties(self) -> Dict[str, bool]:
        """Dictionary of the UserAccountControl properties."""
        return self.__properties

    @property
    def value(self) -> int:
        """The integer value of the properties."""
        return sum(
            self.__flag_values[key] for key, val in self.properties.items() if val
        )


__all__ = [
    "ACE",
    "ACEFlag",
    "ACERight",
    "ACEType",
    "ACL",
    "ACLRevision",
    "SecurityDescriptor",
    "SID",
    "UserAccountControl",
]



import os
import struct
import sys

import pytest
from conftest import get_config, test_ace

from bonsai import LDAPClient
from bonsai.active_directory.acl import ACE, ACL, ACLRevision
from bonsai.active_directory.sid import SID
from bonsai.active_directory import SecurityDescriptor


@pytest.fixture
def client():
    """Create a client with authentication settings."""
    cfg = get_config()
    url = f"ldap://{cfg['SERVER']['hostname']}:{cfg['SERVER']['port']}"
    client = LDAPClient(url)
    client.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    return client


@pytest.fixture
def test_sd():
    return SecurityDescriptor(
        {
            "self_relative": False,
            "rm_control_valid": False,
            "sacl_protected": False,
            "dacl_protected": False,
            "sacl_auto_inherited": False,
            "dacl_auto_inherited": False,
            "sacl_computed_inheritance_required": False,
            "dacl_computed_inheritance_required": False,
            "server_security": False,
            "dacl_trusted": True,
            "sacl_defaulted": False,
            "sacl_present": False,
            "dacl_defaulted": False,
            "dacl_present": False,
            "group_defaulted": True,
            "owner_defaulted": True,
        },
        SID("S-1-1-0"),
        SID("S-1-1-0"),
        None,
        None,
    )


def test_from_binary():
    """Test from_binary method."""
    with pytest.raises(TypeError):
        _ = SecurityDescriptor.from_binary(0)
    with pytest.raises(TypeError):
        _ = SecurityDescriptor.from_binary("INVALID")
    with pytest.raises(ValueError):
        _ = SecurityDescriptor.from_binary(b"\x05\nH\x00\x07\x00")
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", "sd-sample0.bin"), "rb") as data:
        input_data = data.read()
        sec_desc = SecurityDescriptor.from_binary(input_data)
        assert sec_desc.revision == 1
        assert sec_desc.group_sid == "S-1-5-21-3526669579-2242266465-3136906013-512"
        assert sec_desc.owner_sid == "S-1-5-21-3526669579-2242266465-3136906013-512"
        assert sec_desc.sbz1 == 0
        assert sec_desc.control["dacl_present"]
        assert len(sec_desc.dacl.aces) == 24
        assert sec_desc.dacl.aces[0].type == 5
        assert str(sec_desc.dacl.aces[0].trustee_sid) == "S-1-5-32-554"
        assert not sec_desc.control["sacl_present"]
        assert sec_desc.sacl is None
    with open(os.path.join(curdir, "testenv", "sd-sample1.bin"), "rb") as data:
        input_data = data.read()
        sec_desc = SecurityDescriptor.from_binary(input_data)
        assert sec_desc.revision == 1
        assert sec_desc.group_sid == "S-1-5-21-3526669579-2242266465-3136906013-512"
        assert sec_desc.owner_sid == "S-1-5-21-3526669579-2242266465-3136906013-512"
        assert sec_desc.sbz1 == 0
        assert sec_desc.control["dacl_present"]
        assert len(sec_desc.dacl.aces) == 24
        assert sec_desc.dacl.aces[0].type == 5
        assert sec_desc.dacl.aces[0].trustee_sid == "S-1-5-32-554"
        assert sec_desc.control["sacl_present"]
        assert len(sec_desc.sacl.aces) == 3
        assert sec_desc.sacl.aces[0].type == 2
        assert sec_desc.sacl.aces[0].trustee_sid == "S-1-1-0"


@pytest.mark.parametrize(
    "file",
    ["sd-sample0.bin", "sd-sample1.bin"],
)
def test_to_binary(file):
    """Test to_binary method."""
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", file), "rb") as data:
        expected_data = data.read()
        test_sec_desc = SecurityDescriptor.from_binary(expected_data)
        test_data = test_sec_desc.to_binary()
        (
            expected_rev,
            expected_sbz1,
            expected_ctrl,
            expected_offset_owner,
            expected_offset_group,
            expected_offset_sacl,
            expected_offset_dacl,
        ) = struct.unpack("<BBHIIII", expected_data[:20])
        (
            test_rev,
            test_sbz1,
            test_ctrl,
            test_offset_owner,
            test_offset_group,
            test_offset_sacl,
            test_offset_dacl,
        ) = struct.unpack("<BBHIIII", test_data[:20])
        assert len(test_data) == len(expected_data)
        assert test_rev == expected_rev
        assert test_sbz1 == expected_sbz1
        assert test_ctrl == expected_ctrl
        if expected_offset_owner:
            assert (
                test_data[
                    test_offset_owner : test_offset_owner + test_sec_desc.owner_sid.size
                ]
                == expected_data[
                    expected_offset_owner : expected_offset_owner
                    + test_sec_desc.owner_sid.size
                ]
            )
        if expected_offset_group:
            assert (
                test_data[
                    test_offset_group : test_offset_group + test_sec_desc.group_sid.size
                ]
                == expected_data[
                    expected_offset_group : expected_offset_group
                    + test_sec_desc.group_sid.size
                ]
            )
        if expected_offset_sacl:
            assert (
                test_data[test_offset_sacl : test_offset_sacl + test_sec_desc.sacl.size]
                == expected_data[
                    expected_offset_sacl : expected_offset_sacl
                    + test_sec_desc.sacl.size
                ]
            )
        if expected_offset_dacl:
            assert (
                test_data[test_offset_dacl : test_offset_dacl + test_sec_desc.dacl.size]
                == expected_data[
                    expected_offset_dacl : expected_offset_dacl
                    + test_sec_desc.dacl.size
                ]
            )
        assert SecurityDescriptor.from_binary(test_data).to_binary() == test_data


@pytest.mark.skipif(
    not sys.platform.startswith("win"),
    reason="Cannot query SecurityDescriptor from OpenLDAP",
)
@pytest.mark.parametrize(
    "sd_flags, owner_sid, group_sid, dacl, sacl",
    [
        (1, True, False, False, False),
        (2, False, True, False, False),
        (3, True, True, False, False),
        (4, False, False, True, False),
        (8, False, False, False, True),
        (15, True, True, True, True),
    ],
    ids=["only-owner", "only-group", "owner-group", "only-dacl", "only-sacl", "all"],
)
def test_sd_flags(client, sd_flags, owner_sid, group_sid, dacl, sacl):
    """Test LDAP_SERVER_SD_FLAGS_OID control"""
    client.sd_flags = sd_flags
    with client.connect() as conn:
        res = conn.search(
            "cn=chuck,ou=nerdherd,dc=bonsai,dc=test",
            0,
            attrlist=["nTSecurityDescriptor"],
        )[0]
        sec_desc = SecurityDescriptor.from_binary(res["nTSecurityDescriptor"][0])
        assert sec_desc.revision == 1
        if owner_sid:
            assert sec_desc.owner_sid is not None
            assert isinstance(sec_desc.owner_sid, SID)
        else:
            assert sec_desc.owner_sid is None
        if group_sid:
            assert sec_desc.group_sid is not None
            assert isinstance(sec_desc.group_sid, SID)
        else:
            assert sec_desc.group_sid is None
        assert sec_desc.control["dacl_present"] is dacl
        if dacl:
            assert isinstance(sec_desc.dacl, ACL)
        else:
            assert sec_desc.dacl is None
        assert sec_desc.control["sacl_present"] is sacl
        if sacl:
            assert isinstance(sec_desc.sacl, ACL)
        else:
            assert sec_desc.sacl is None


def test_sd_set_control(test_sd: SecurityDescriptor):
    """Test setting controls for SecurityDescriptor."""
    with pytest.raises(TypeError):
        test_sd.set_control("string")
    with pytest.raises(TypeError):
        test_sd.set_control({"a": "b", "c": True})
    with pytest.raises(ValueError):
        test_sd.set_control(-2)
    test_sd.set_control(0)
    assert all(val is False for val in test_sd.control.values())
    test_sd.control = {"sacl_protected": True}
    assert test_sd.control["sacl_protected"]
    assert all(
        val is False for key, val in test_sd.control.items() if key != "sacl_protected"
    )
    test_sd.control = 0x8 + 0x20
    assert test_sd.control["dacl_defaulted"]
    assert test_sd.control["sacl_defaulted"]
    assert all(
        val is False
        for key, val in test_sd.control.items()
        if key not in ("dacl_defaulted", "sacl_defaulted")
    )


def test_sd_set_owner_sid(test_sd: SecurityDescriptor):
    """Test setting owner SID for SecurityDescriptor."""
    assert test_sd.owner_sid == SID("S-1-1-0")
    with pytest.raises(TypeError):
        test_sd.set_owner_sid("string")
    with pytest.raises(TypeError):
        test_sd.set_owner_sid(0)
    with pytest.raises(TypeError):
        test_sd.owner_sid = True
    test_sd.set_owner_sid(SID("S-1-5-10"))
    assert test_sd.owner_sid == SID("S-1-5-10")
    test_sd.owner_sid = SID("S-1-5-20")
    assert test_sd.owner_sid == SID("S-1-5-20")


def test_sd_set_group_sid(test_sd: SecurityDescriptor):
    """Test setting group SID for SecurityDescriptor."""
    assert test_sd.group_sid == SID("S-1-1-0")
    with pytest.raises(TypeError):
        test_sd.set_group_sid("string")
    with pytest.raises(TypeError):
        test_sd.set_group_sid(0)
    with pytest.raises(TypeError):
        test_sd.group_sid = True
    test_sd.set_group_sid(SID("S-1-5-10"))
    assert test_sd.group_sid == SID("S-1-5-10")
    test_sd.group_sid = SID("S-1-5-20")
    assert test_sd.group_sid == SID("S-1-5-20")


def test_sd_set_dacl(test_sd: SecurityDescriptor, test_ace: ACE):
    """Test setting dacl for SecurityDescriptor."""
    assert test_sd.dacl is None
    with pytest.raises(TypeError):
        test_sd.set_dacl("string")
    with pytest.raises(TypeError):
        test_sd.set_dacl(0)
    with pytest.raises(TypeError):
        test_sd.dacl = True
    with pytest.raises(TypeError):
        test_sd.set_dacl(["A", "b"])
    test_sd.set_dacl(ACL(ACLRevision.ACL_REVISION, []))
    assert test_sd.dacl is not None
    assert test_sd.control["dacl_present"]
    test_sd.dacl = ACL(ACLRevision.ACL_REVISION, [test_ace])
    assert test_sd.dacl is not None
    assert test_sd.dacl.aces[0] == test_ace
    test_sd.dacl = None
    assert test_sd.dacl is None
    assert test_sd.control["dacl_present"] is False


def test_sd_set_sacl(test_sd: SecurityDescriptor, test_ace: ACE):
    """Test setting sacl for SecurityDescriptor."""
    assert test_sd.sacl is None
    with pytest.raises(TypeError):
        test_sd.set_sacl("string")
    with pytest.raises(TypeError):
        test_sd.set_sacl(0)
    with pytest.raises(TypeError):
        test_sd.sacl = True
    with pytest.raises(TypeError):
        test_sd.set_sacl(["A", "b"])
    test_sd.set_sacl(ACL(ACLRevision.ACL_REVISION, []))
    assert test_sd.sacl is not None
    assert test_sd.control["sacl_present"]
    test_sd.sacl = ACL(ACLRevision.ACL_REVISION, [test_ace])
    assert test_sd.sacl is not None
    assert test_sd.sacl.aces[0] == test_ace
    test_sd.sacl = None
    assert test_sd.sacl is None
    assert test_sd.control["sacl_present"] is False



import os
import uuid

import pytest

from bonsai.active_directory import ACL, SID
from bonsai.active_directory.acl import ACE, ACEFlag, ACEType, ACERight, ACLRevision

from conftest import test_ace

ACE_BINARY_TEST_DATA = b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_(\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\x15\xb3LK\xd2\xfb\x90s\xc2\xdf9\xb9\\\x04\x00\x00"


def test_ace_from_binary():
    """Test ACE's from_binary method."""
    with pytest.raises(TypeError):
        _ = ACE.from_binary(0)
    with pytest.raises(TypeError):
        _ = ACE.from_binary("INVALID")
    with pytest.raises(ValueError):
        _ = ACE.from_binary(b"\x05\nH\x00\x07\x00\x00\x00")
    ace = ACE.from_binary(ACE_BINARY_TEST_DATA)
    assert ace.type == ACEType.ACCESS_ALLOWED_OBJECT
    assert ace.type.is_object_type
    assert ace.type.is_access_allowed
    assert not ace.type.is_access_denied
    assert ace.mask == 7
    assert ace.rights == {
        ACERight.DS_CREATE_CHILD,
        ACERight.DS_DELETE_CHILD,
        ACERight.ACTRL_DS_LIST,
    }
    assert ace.size == len(ACE_BINARY_TEST_DATA)
    assert ace.flags == {ACEFlag.INHERIT_ONLY, ACEFlag.CONTAINER_INHERIT}
    assert ace.is_inherited is False
    assert sum(ace.flags) == 10
    assert ace.trustee_sid == "S-1-5-21-1263317781-1938881490-3107577794-1116"
    assert ace.object_type == uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506")
    assert ace.inherited_object_type == uuid.UUID(
        "4828cc14-1437-45bc-9b07-ad6f015e5f28"
    )
    assert ace.application_data == b""


def test_ace_to_binary():
    """Test ACE's from_binary method."""
    assert ACE.from_binary(ACE_BINARY_TEST_DATA).to_binary() == ACE_BINARY_TEST_DATA


def test_ace_set_access_rights(test_ace: ACE):
    """Test setting ACE's access rights."""
    assert test_ace.rights == {ACERight.GENERIC_READ}
    with pytest.raises(TypeError):
        test_ace.set_access_rights(None)
    with pytest.raises(TypeError):
        test_ace.set_access_rights("GENERIC_READ")
    with pytest.raises(TypeError):
        test_ace.set_access_rights({ACERight.GENERIC_READ, 12})
    with pytest.raises(ValueError):
        test_ace.set_access_rights(-12)
    test_ace.set_access_rights(268435464)
    assert test_ace.mask == 268435464
    assert test_ace.rights == {ACERight.GENERIC_ALL, ACERight.DS_SELF}
    test_ace.mask = 0
    assert test_ace.rights == set()
    new_rights = {ACERight.GENERIC_WRITE, ACERight.ACTRL_DS_LIST}
    test_ace.rights = new_rights
    assert test_ace.rights == new_rights
    assert test_ace.mask == sum(new_rights)


def test_ace_set_application_data(test_ace: ACE):
    """Test setting ACE's application data."""
    assert test_ace.application_data == b""
    with pytest.raises(TypeError):
        test_ace.set_application_data("string")
    test_ace.set_application_data(b"\x00\x10")
    assert test_ace.application_data == b"\x00\x10"
    test_ace.application_data = b"\x11\x10"
    assert test_ace.application_data == b"\x11\x10"


def test_ace_set_flags(test_ace: ACE):
    """Test setting ACE's flags."""
    assert test_ace.flags == {ACEFlag.INHERITED}
    with pytest.raises(TypeError):
        test_ace.set_flags("string")
    with pytest.raises(TypeError):
        test_ace.set_flags(42)
    test_ace.set_flags({ACEFlag.INHERIT_ONLY})
    assert test_ace.flags == {ACEFlag.INHERIT_ONLY}
    test_ace.flags = {ACEFlag.OBJECT_INHERIT, ACEFlag.SUCCESSFUL_ACCESS}
    assert test_ace.flags == {ACEFlag.OBJECT_INHERIT, ACEFlag.SUCCESSFUL_ACCESS}


def test_ace_set_object_type(test_ace: ACE):
    """Test setting ACE's object type."""
    assert test_ace.object_type is None
    with pytest.raises(TypeError):
        test_ace.set_object_type("string")
    with pytest.raises(TypeError):
        test_ace.set_object_type(42)
    obj_id = uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506")
    test_ace.set_object_type(obj_id)
    assert test_ace.object_type == obj_id
    test_ace.object_type = None
    assert test_ace.object_type is None


def test_ace_set_inherited_object_type(test_ace: ACE):
    """Test setting ACE's inherited object type."""
    assert test_ace.inherited_object_type == uuid.UUID(
        "c975c901-6cea-4b6f-8319-d67f45449506"
    )
    with pytest.raises(TypeError):
        test_ace.set_inherited_object_type("string")
    with pytest.raises(TypeError):
        test_ace.set_inherited_object_type(42)
    obj_id = uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449507")
    test_ace.set_inherited_object_type(obj_id)
    assert test_ace.inherited_object_type == obj_id
    test_ace.inherited_object_type = None
    assert test_ace.inherited_object_type is None


def test_ace_set_type(test_ace: ACE):
    """Test setting ACE's type."""
    assert test_ace.type == ACEType.ACCESS_ALLOWED
    assert test_ace.type.is_access_allowed
    with pytest.raises(TypeError):
        test_ace.set_type("string")
    with pytest.raises(TypeError):
        test_ace.set_type(42)
    test_ace.set_type(ACEType.ACCESS_DENIED)
    assert test_ace.type == ACEType.ACCESS_DENIED
    assert test_ace.type.is_access_denied
    test_ace.type = ACEType.ACCESS_DENIED_CALLBACK
    assert test_ace.type == ACEType.ACCESS_DENIED_CALLBACK
    assert test_ace.type.is_access_denied


def test_ace_set_trustee_sid(test_ace: ACE):
    """Test setting ACE's trustee SID."""
    assert test_ace.trustee_sid == SID("S-1-1-0")
    with pytest.raises(TypeError):
        test_ace.set_trustee_sid("string")
    with pytest.raises(TypeError):
        test_ace.set_trustee_sid(42)
    test_ace.set_trustee_sid(SID("S-1-5-10"))
    assert test_ace.trustee_sid == SID("S-1-5-10")
    test_ace.trustee_sid = SID("S-1-5-11")
    assert test_ace.trustee_sid == SID("S-1-5-11")


def test_eq(test_ace: ACE):
    """Test ACE's __eq__ method."""
    assert not test_ace == 21
    other_ace = ACE(
        ACEType.ACCESS_ALLOWED,
        {ACEFlag.INHERITED},
        0x10000000,
        SID("S-1-1-0"),
        None,
        uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506"),
        b"",
    )
    assert not test_ace == other_ace
    other_ace = ACE(
        ACEType.ACCESS_ALLOWED,
        {ACEFlag.INHERITED},
        0x80000000,
        SID("S-1-1-0"),
        None,
        uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506"),
        b"\x01",
    )
    assert not test_ace == other_ace
    other_ace = ACE.from_binary(ACE_BINARY_TEST_DATA)
    assert not test_ace == other_ace
    other_ace = ACE(
        ACEType.ACCESS_ALLOWED,
        {ACEFlag.INHERITED},
        0x80000000,
        SID("S-1-1-0"),
        None,
        uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506"),
        b"",
    )
    assert test_ace == other_ace


def test_acl_from_binary():
    """Test ACL's from_binary method."""
    with pytest.raises(TypeError):
        _ = ACL.from_binary(True)
    with pytest.raises(TypeError):
        _ = ACL.from_binary("INVALID")
    with pytest.raises(ValueError):
        _ = ACL.from_binary(
            b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_("
        )
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", "acl-sample.bin"), "rb") as data:
        input_data = data.read()
        acl = ACL.from_binary(input_data)
        assert acl.size == len(input_data)
        assert acl.revision == ACLRevision.ACL_REVISION_DS
        assert len(acl.aces) == 85
        assert all(isinstance(ace, ACE) for ace in acl.aces)
        assert acl.aces[0].type == ACEType.ACCESS_ALLOWED_OBJECT
        assert acl.aces[0].inherited_object_type == uuid.UUID(
            "4828cc14-1437-45bc-9b07-ad6f015e5f28"
        )
        assert acl.aces[-1].type == ACEType.ACCESS_ALLOWED
        assert (
            acl.aces[0].trustee_sid == "S-1-5-21-1263317781-1938881490-3107577794-1116"
        )
        assert acl.aces[-1].trustee_sid == "S-1-5-18"


def test_acl_to_binary():
    """Test ACL's from_binary method."""
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", "acl-sample.bin"), "rb") as data:
        input_data = data.read()
        assert ACL.from_binary(input_data).to_binary() == input_data


def test_acl_set_aces(test_ace):
    """Test setting ACL's aces list."""
    test_acl = ACL(ACLRevision.ACL_REVISION, [])
    assert test_acl.aces == []
    with pytest.raises(TypeError):
        test_acl.set_aces([1, 2, 3])
    with pytest.raises(TypeError):
        test_acl.set_aces([test_ace, 3])
    test_acl.set_aces([test_ace])
    assert test_acl.aces == [test_ace]
    test_acl.aces = []
    assert test_acl.aces == []


def test_str():
    """Test __str__ method."""
    ace = ACE.from_binary(ACE_BINARY_TEST_DATA)
    assert (
        str(ace)
        == "(OA;CIIO;CCDCLC;c975c901-6cea-4b6f-8319-d67f45449506;4828cc14-1437-45bc-9b07-ad6f015e5f28;S-1-5-21-1263317781-1938881490-3107577794-1116)"
    )
